# """check the weather"""
# from datetime import datetime, timedelta
# import uuid
#
# from apscheduler.schedulers.asyncio import AsyncIOScheduler  # type: ignore
#
# from wechaty_puppet import get_logger
#
# from wechaty import Message, Wechaty
# from wechaty.plugin import WechatyPlugin
#
#
# log = get_logger('DingDongPluginBot')
#
#
# class DingDongRankPlugin(WechatyPlugin):
#     """weather plugin for bot"""
#
#     def __init__(self):
#         """init the plugin"""
#         super().__init__()
#
#         self.room_data = {}
#         self.interval_time = 8
#         self.clear_seconds = 5
#
#         self.top_k = 10
#         self.room_alias = {}
#
#         self.scheduler = AsyncIOScheduler()
#         self.scheduler.start()
#
#     @property
#     def name(self) -> str:
#         """get the name of the plugin"""
#         return 'ding-dong-rank'
#
#     async def on_message(self, msg: Message):
#         """listen message event"""
#         from_contact = msg.talker()
#         text = msg.text()
#         room = msg.room()
#
#         if room is None:
#             return
#
#         if text in ['#ding', 'dong']:
#             if room.room_id not in self.room_data:
#                 self.room_data[room.room_id] = {
#                     '__status__': 'running'
#                 }
#                 await self.re_init_job(room.room_id)
#         if text == '#ding':
#             self.room_data[room.room_id]['__ding__'] = True
#
#         if text == '#ding start':
#             self.room_data[room.room_id] = {
#                 '__status__': 'running'
#             }
#             await room.say('#ding')
#             await self.re_init_job(room.room_id)
#
#         if text == 'dong':
#             if room.room_id in self.room_data:
#                 if self.room_data[room.room_id]['__status__'] != 'running':
#                     return
#
#                 self.room_data[room.room_id][from_contact.contact_id] = {
#                     'id': from_contact.contact_id,
#                     'time': datetime.now(),
#                     'name': from_contact.name,
#                 }
#                 # refresh the scheduler job
#                 await self.re_init_job(room.room_id)
#
#     async def get_room_alias(self, room_id, member_id, name):
#         """get alias in room
#             this function is not supported in donut wechaty-puppet
#         """
#         room = self.bot.Room(room_id)
#         await room.ready()
#         members = await room.member_all()
#
#         if room_id not in self.room_alias:
#             self.room_alias[room_id] = {}
#             for member in members:
#                 await member.ready()
#                 alias = await room.alias(member)
#                 if alias is None or alias == '':
#                     self.room_alias[room_id][member.contact_id] = \
#                         member.payload.name
#                 else:
#                     self.room_alias[room_id][member.contact_id] = alias
#
#         if member_id not in self.room_alias[room_id]:
#             return name
#         return self.room_alias[room_id][member_id]
#
#     async def send_rank_analysis(self, room_id):
#         """send rank analysis data to conversation room"""
#         log.info('send rank analysis ...')
#         if room_id not in self.room_data:
#             return
#
#         # check if contains ding info
#         if not self.room_data[room_id].get('__ding__', False):
#             return
#
#         bot_ids = [key for key in self.room_data[room_id].keys() if
#                    not key.startswith('__')]
#         dong_data = [self.room_data[room_id][bot_id] for bot_id in bot_ids]
#
#         # sort the date
#         print(dong_data)
#         dong_data.sort(key=lambda x: x['time'])
#
#         top_k_result = []
#         last_timestamp = datetime(year=2000, month=1, day=1)
#         top = 0
#
#         ding_dong_histories = []
#
#         for index, item in enumerate(dong_data):
#
#             if index <= self.top_k:
#                 if item['time'] > last_timestamp:
#                     last_timestamp = item['time']
#                     top += 1
#                 top_k_result.append({
#                     'id': item['id'],
#                     'name': item['name'],
#                     'top': top
#                 })
#                 ding_dong_histories.append(
#                     DingDongHistory(
#                         id=str(uuid.uuid4()),
#                         contact_id=item['id'],
#                         time=item['time'],
#                         room_id=room_id,
#                         rank=top
#                     )
#                 )
#
#         # save history to db
#
#         # send message to the room
#         msg = '🔥🔥 DING-DONG Speed Ranking \n=========================\n'
#
#         icon = {
#             1: '1️⃣', 2: '2️⃣', 3: '3️⃣', 4: '4️⃣', 5: '5️⃣',
#             6: '6️⃣', 7: '7️⃣', 8: '8️⃣', 9: '9️⃣', 10: '🔟'
#         }
#
#         for item in top_k_result:
#             if item['top'] in icon:
#                 msg += f'NO.{icon[item["top"]]} {item["name"]}\n'
#
#         room = self.bot.Room.load(room_id)
#         await room.ready()
#         await room.say(msg)
#
#         del self.room_data[room_id]
#
#     def stop_ding_dong_rank(self, room_id: str):
#         """clear all ding-dong history"""
#         log.info('clear ding dong history ...')
#         if room_id in self.room_data:
#             self.room_data[room_id]['__status__'] = 'stopping'
#
#     async def re_init_job(self, room_id):
#         """re init job, replace it if exist"""
#         job_id = f'__job__{room_id}'
#         now = datetime.now()
#
#         # clear all response data in 3 seconds
#         clear_job = self.scheduler.add_job(
#             self.stop_ding_dong_rank,
#             kwargs={'room_id': room_id},
#             id=f'__clear_history__{room_id}',
#             trigger='date',
#             next_run_time=now + timedelta(seconds=self.clear_seconds),
#             replace_existing=True
#         )
#         log.info(clear_job)
#
#         rank_job = self.scheduler.add_job(
#             self.send_rank_analysis,
#             replace_existing=True,
#             id=job_id,
#             trigger='date',
#             next_run_time=now + timedelta(seconds=self.interval_time),
#             kwargs={'room_id': room_id}
#         )
#         log.info(rank_job)



"""
python-wechaty-plugin-contrib - https://github.com/wechaty/python-wechaty-plugin-contrib

Authors:    Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright wj-Mcat

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from dataclasses import (
    dataclass,
    field
)
from typing import Optional, List

from wechaty import (
    WechatyPluginOptions
)


@dataclass
class GitlabHookItem:
    """plugin can hook many gitlab projects, so the GitlabHookItem can specific
    the project
    """
    id: str

    room_topic: Optional[str] = None
    room_id: Optional[str] = None

    contact_name: Optional[str] = None
    contact_id: Optional[str] = None


@dataclass
class GitlabEventOptions(WechatyPluginOptions):
    hook_url: Optional[str] = None
    secret_token: Optional[str] = None

    listen_port: Optional[int] = 5100

    hook_items: List[GitlabHookItem] = field(default_factory=list)



"""
python-wechaty-plugin-contrib - https://github.com/wechaty/python-wechaty-plugin-contrib

Authors:    Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright wj-Mcat

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations
import asyncio
from wechaty import (
    Wechaty, WechatyPlugin,
)

from aiohttp import web
from wechaty_puppet import RoomQueryFilter, get_logger

from .options import GitlabHookItem, GitlabEventOptions

log = get_logger("GitlabEventPlugin")
routes = web.RouteTableDef()


@routes.get('/')
async def receive_message(request: web.Request):
    """"""
    log.info('')
    bot: Wechaty = request.app.get('bot')
    options: GitlabEventOptions = request.app.get('bot-options')

    return web.json_response(text='hello wechaty web bot')


class GitlabEventPlugin(WechatyPlugin):
    """"""

    def __init__(self, options: GitlabEventOptions):
        super().__init__(options)
        self.options = options

    @property
    def name(self) -> str:
        return 'gitlab-event-plugin'

    async def init_plugin(self, wechaty: Wechaty):
        """init the gitlab event plugin"""
        log.info('starting the server')
        self.bot = wechaty

        # start the server
        app = web.Application()
        app['bot'] = wechaty
        app['bot-options'] = self.options
        app.add_routes(routes)

        runner = web.AppRunner(app)
        await runner.setup()

        site = web.TCPSite(runner, '0.0.0.0', self.options.listen_port)

        loop = asyncio.get_event_loop()
        asyncio.run_coroutine_threadsafe(site.start(), loop=loop)
        log.info(
            f'the server has started ... 0.0.0.0: {self.options.listen_port}'
        )



"""AutoReply to someone according to keywords"""
from dataclasses import dataclass, field
from datetime import datetime
from collections import defaultdict
from typing import Union, List, Dict, Optional

from wechaty import (  # type: ignore
    WechatyPlugin,
    WechatyPluginOptions,
    FileBox,
    Contact,
    Message, Room, Wechaty
)

from wechaty_puppet import (    # type: ignore
    get_logger,
)

from wechaty_plugin_contrib.matchers import (
    Matcher,
    MatcherOptions,
    MatcherOption, RoomMatcher, MessageMatcher, ContactMatcher
)

from wechaty_plugin_contrib.exception import (
    WechatyPluginConfigurationError
)


@dataclass
class RoomInviterOptions(WechatyPluginOptions):
    # add rules to the specific Matcher
    rules: Dict[MessageMatcher, Union[List[RoomMatcher], RoomMatcher]] = field(default_factory=dict)
    welcome: str = field(default_factory=str)


logger = get_logger('RoomInviterPlugin')


class RoomInviterPlugin(WechatyPlugin):

    def __init__(self, options: RoomInviterOptions):
        super().__init__(options)
        if not options or not options.rules:
            raise ValueError('options is required, please add matcher rules')

        self.rules: Dict[MessageMatcher, List[RoomMatcher]] = {}
        for matcher, room_obj in options.rules.items():
            if not room_obj:
                raise ValueError(f'matcher<{matcher}> can not match the target room')
            if isinstance(room_obj, RoomMatcher):
                self.rules[matcher] = [room_obj]
            elif isinstance(room_obj, list):
                self.rules[matcher] = room_obj
            else:
                raise ValueError(f'configuration is error')
        self.welcome_ids: Dict[str, List[str]] = defaultdict(list)
        self.welcome_words = options.welcome

    async def init_plugin(self, wechaty: Wechaty):
        """listen room-join events and say welcome"""
        async def on_room_join(room: Room, invitees: List[Contact], *args):
            if room.room_id in self.welcome_ids:
                for invited_contact in invitees:
                    if invited_contact.contact_id in self.welcome_ids[room.room_id]:
                        await room.say(
                            self.welcome_words,
                            mention_ids=[invited_contact.contact_id]
                        )
                        self.welcome_ids[room.room_id].remove(
                            invited_contact.contact_id
                        )

        wechaty.on('room-join', on_room_join)

    async def on_message(self, msg: Message):
        """check the keyword and reply to talker"""

        # don't listen in room
        if msg.room():
            return

        talker = msg.talker()

        for message_matcher, rules in self.rules.items():
            is_match = await message_matcher.match(msg)

            if is_match:
                # cache the rooms data
                rooms = []
                for rule in rules:
                    # pytype: disable=attribute-error
                    matched_rooms = await rule.find_rooms(self.bot)
                    rooms.extend(matched_rooms)

                # invite person to the room
                for room in rooms:
                    self.welcome_ids[room.room_id].append(talker.contact_id)
                    await room.add(talker)



"""basic ding-dong bot for the wechaty plugin"""
from __future__ import annotations
from typing import List, Optional, Union
from dataclasses import dataclass, field
import requests

from wechaty import Message, Contact, Room, get_logger  # type: ignore
from wechaty.exceptions import WechatyPluginError   # type: ignore
from wechaty.plugin import WechatyPlugin, WechatyPluginOptions  # type: ignore
from wechaty_grpc.wechaty.puppet import MessageType     # type: ignore

# from wechaty_plugin_contrib.finders.finder import Finder  # type: ignore


log = get_logger('DingDongPlugin')


@dataclass
class RasaRestPluginOptions(WechatyPluginOptions):
    """Rasa Rest Connector Options
    """
    endpoint: Optional[str] = None
    conversation_ids: List[str] = field(default_factory=list)


class RasaRestPlugin(WechatyPlugin):
    """rasa server channnel connector"""
    def __init__(self, options: Optional[RasaRestPluginOptions] = None):
        super().__init__(options)
        if not options or not options.endpoint:
            raise WechatyPluginError(
                'please set the endpoint config to RasaRestPlugin, '
                'you can use RasaRestPluginOptions or kwargs'
            )
        endpoint = options.endpoint
        if not endpoint.endswith('webhooks/rest/webhook'):
            if not endpoint.endswith('/'):
                endpoint += '/'

            endpoint = f'{endpoint}webhooks/rest/webhook'
        self.endpoint = endpoint
        self.conversation_ids: List[str] = options.conversation_ids if options else []

    @property
    def name(self):
        """name of the plugin"""
        return 'rasa-rest-plugin'

    async def on_message(self, msg: Message):
        """listen message event"""
        talker = msg.talker()
        room = msg.room()

        conversation_id = room.room_id if room else talker.contact_id

        # 1. TODO: use finder to match Room/Contact
        if conversation_id not in self.conversation_ids:
            return

        # only process the plain text message
        if msg.type() != MessageType.MESSAGE_TYPE_TEXT:
            return

        rasa_response = requests.post(
            self.endpoint,
            json=dict(
                sender=conversation_id,
                message=msg.text()
            )
        )
        messages: List[dict] = rasa_response.json()
        if len(messages) == 0:
            return

        conversational: Union[Room, Contact] = room if room else talker
        for message in messages:
            msg_text = message.get('text', None)
            if not msg_text:
                continue
            await conversational.say(msg_text)



"""
utils to help plugins more stronger
"""
import re


def is_regex_pattern(pattern: str) -> bool:
    """check if the string is a valid regex pattern"""
    try:
        re.compile(pattern)
        is_valid = True
    except re.error:
        is_valid = False
    return is_valid



"""Room Finder to match the specific Room"""
import re
from re import Pattern
import inspect
from typing import List

from wechaty_plugin_contrib.config import (
    get_logger,
    Room,

    Wechaty
)

from .finder import Finder


logger = get_logger("RoomFinder")


class RoomFinder(Finder):
    async def match(self, wechaty: Wechaty) -> List[Room]:
        """match the room"""
        logger.info(f'RoomFinder match({Wechaty})')

        rooms: List[Room] = []

        for option in self.options:
            if isinstance(option, Pattern):

                # search from all of the friends
                # match the room with regex pattern
                all_rooms = await wechaty.Room.find_all()
                for room in all_rooms:
                    topic = await room.topic()
                    if re.match(option, topic):
                        rooms.append(room)

            elif isinstance(option, str):
                room = wechaty.Room.load(option)
                await room.ready()
                rooms.append(room)
            elif hasattr(option, '__call__'):
                """check the type of the function
                refer: https://stackoverflow.com/a/56240578/6894382
                """
                if inspect.iscoroutinefunction(option):
                    # pytype: disable=bad-return-type
                    targets = await option(wechaty)
                else:
                    targets = option(wechaty)
                if isinstance(targets, List[Room]):
                    rooms.extend(targets)
            else:
                raise ValueError(f'unknown type option: {option}')
        return rooms



"""basic ding-dong bot for the wechaty plugin"""
from typing import List, Optional, Union
from dataclasses import dataclass, field
from wechaty import Message, Contact, Room, get_logger  # type: ignore
from wechaty.plugin import WechatyPlugin, WechatyPluginOptions  # type: ignore


log = get_logger('DingDongPlugin')


@dataclass
class DingDongPluginOptions(WechatyPluginOptions):
    """
    conversation_id can be contact_id and room_id.

    only one of [include_conversation_ids,exclude_conversation_ids] can be empty
    """
    include_conversation_ids: List[str] = field(default_factory=list)
    exclude_conversation_ids: List[str] = field(default_factory=list)


class DingDongPlugin(WechatyPlugin):
    """basic ding-dong plugin"""
    def __init__(self, options: Optional[DingDongPluginOptions] = None):
        super().__init__(options)

        if options is not None:
            if options.include_conversation_ids is not None and \
                    options.exclude_conversation_ids is not None:
                log.info('only one of [include_conversation_ids, '
                         'exclude_conversation_ids] can be empty ')

        self.options = options

    @property
    def name(self):
        """name of the plugin"""
        return 'ding-dong'

    def can_send_dong(self, conversation_id: str):
        """check if the bot can send dong message"""
        if self.options is None:
            return True
        if self.options.include_conversation_ids is not None:
            if not isinstance(self.options.include_conversation_ids, list):
                log.error('include_conversation_ids should be a list')
            elif conversation_id in self.options.include_conversation_ids:
                return True
        if self.options.exclude_conversation_ids is not None:
            if not isinstance(self.options.exclude_conversation_ids, list):
                log.error('exclude_conversation_ids should be a list')
            elif conversation_id not in self.options.exclude_conversation_ids:
                return True
        return False

    async def on_message(self, msg: Message):
        """listen message event"""
        from_contact = msg.talker()
        text = msg.text()
        room = msg.room()
        if text == '#ding':
            conversation: Union[Room, Contact] = from_contact if room is None else room
            conversation_id = from_contact.contact_id if room is None \
                else room.room_id
            if self.can_send_dong(conversation_id):
                await conversation.ready()
                await conversation.say('dong')



"""Base finder for finding Room/Message"""
from re import Pattern
from typing import (
    Callable,
    Optional,
    Union,
    List,
)

from wechaty_plugin_contrib.config import (
    Room,
    Contact,
    Message
)


FinderOption = Union[str, Pattern, bool, Callable[[Union[Contact, Room, Message]], List[Union[Room, Contact]]]]
FinderOptions = Union[FinderOption, List[FinderOption]]


class Finder:
    def __init__(self, option: FinderOptions):
        if isinstance(option, list):
            self.options = option
        else:
            self.options = [option]

    def match(self, target) -> bool:
        raise NotImplementedError



# """
# python-wechaty-plugin-contrib - https://github.com/wechaty/python-wechaty-plugin-contrib
#
# Authors:    Jingjing WU (吴京京) <https://github.com/wj-Mcat>
#
# 2020-now @ Copyright wj-Mcat
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# """
# from __future__ import annotations
#
# import asyncio
#
# from aiohttp.web_response import StreamResponse
# from wechaty_puppet import get_logger
# from wechaty import WechatyPlugin, Wechaty
#
# from aiohttp import web, ClientRequest
# from aiohttp.web_server import BaseRequest
#
# from .options import (
#     GithubHookItem,
#     GithubWebhookOptions,
#     GithubContentType
# )
#
# log = get_logger('GithubWebhookPlugin')
#
# import hashlib
# import hmac
# import json
# import six
#
#
# class GithubWebhookPlugin(WechatyPlugin):
#     """
#     receive the github webhook events with the inner work
#     """
#
#     @property
#     def name(self) -> str:
#         return 'github-webhook-plugin'
#
#     def __init__(self, options: GithubWebhookOptions):
#         """
#         init the github webhook plugin
#         """
#         super().__init__(options)
#         log.info(f'init the github-webhook-plugin <{options}>')
#
#         if len(options.hook_items) == 0:
#             raise Exception('the hook_items is expected, please make sure this field is more than one')
#         for hook_item in options.hook_items:
#             if not hook_item.room_id and not hook_item.room_topic:
#                 raise Exception(
#                     f'when you configure project <{hook_item.project_id}>, '
#                     f'<room_id> field or <room_topic> field is expected')
#             if not hook_item.contact_name and not hook_item.contact_id:
#                 raise Exception(
#                     f'when you configure project <{hook_item.project_id}>, '
#                     f'<contact_id> field or <contact_name> field is expected')
#             if not hook_item.secret_token:
#                 raise Exception('the secret_token is expected, please make sure this filed is not None')
#             if isinstance(hook_item.secret_token, bytes):
#                 hook_item.secret_token = hook_item.secret_token.decode('utf-8')
#         self.options: GithubWebhookOptions = options
#
#     async def init_plugin(self, wechaty: Wechaty):
#         """
#         init the github-webhook channel with aiohttp web framework, which can listen multi-github projects
#         """
#         log.info(f'init the plugin :{self.name}')
#         app = web.Application()
#
#         async def handle_request(request: StreamResponse):
#             print('================')
#             print(request)
#             print('================')
#
#         app.add_routes([
#             web.get('/', handle_request)
#         ])
#
#         runner = web.AppRunner(app)
#         await runner.setup()
#
#         site = web.TCPSite(runner, '0.0.0.0', self.options.listen_port)
#         loop = asyncio.get_event_loop()
#         asyncio.run_coroutine_threadsafe(site.start(), loop)
#
#         log.info(
#             f'github-webhook plugin server has started at '
#             f'0.0.0.0:{self.options.listen_port}'
#         )
#
#     def hook(self, event_type="push"):
#         """
#         Registers a function as a hook. Multiple hooks can be registered for a given type, but the
#         order in which they are invoke is unspecified.
#         :param event_type: The event type this hook will be invoked for.
#         """
#
#         def decorator(func):
#             self._hooks[event_type].append(func)
#             return func
#
#         return decorator
#
#     def _get_digest(self):
#         """Return message digest if a secret key was provided"""
#
#         return hmac.new(self._secret, request.data, hashlib.sha1).hexdigest() if self._secret else None
#
#     def _postreceive(self):
#         """Callback from Flask"""
#
#         digest = self._get_digest()
#
#         if digest is not None:
#             sig_parts = _get_header("X-Hub-Signature").split("=", 1)
#             if not isinstance(digest, six.text_type):
#                 digest = six.text_type(digest)
#
#             if len(sig_parts) < 2 or sig_parts[0] != "sha1" or not hmac.compare_digest(sig_parts[1], digest):
#                 abort(400, "Invalid signature")
#
#         event_type = _get_header("X-Github-Event")
#         content_type = _get_header("content-type")
#         data = (
#             json.loads(request.form.to_dict(flat=True)["payload"])
#             if content_type == "application/x-www-form-urlencoded"
#             else request.get_json()
#         )
#
#         if data is None:
#             abort(400, "Request body must contain json")
#
#         self._logger.info("%s (%s)", _format_event(event_type, data), _get_header("X-Github-Delivery"))
#
#         for hook in self._hooks.get(event_type, []):
#             hook(data)
#
#         return "", 204
#
#
# EVENT_DESCRIPTIONS = {
#     "commit_comment": "{comment[user][login]} commented on " "{comment[commit_id]} in {repository[full_name]}",
#     "create": "{sender[login]} created {ref_type} ({ref}) in " "{repository[full_name]}",
#     "delete": "{sender[login]} deleted {ref_type} ({ref}) in " "{repository[full_name]}",
#     "deployment": "{sender[login]} deployed {deployment[ref]} to "
#     "{deployment[environment]} in {repository[full_name]}",
#     "deployment_status": "deployment of {deployement[ref]} to "
#     "{deployment[environment]} "
#     "{deployment_status[state]} in "
#     "{repository[full_name]}",
#     "fork": "{forkee[owner][login]} forked {forkee[name]}",
#     "gollum": "{sender[login]} edited wiki pages in {repository[full_name]}",
#     "issue_comment": "{sender[login]} commented on issue #{issue[number]} " "in {repository[full_name]}",
#     "issues": "{sender[login]} {action} issue #{issue[number]} in " "{repository[full_name]}",
#     "member": "{sender[login]} {action} member {member[login]} in " "{repository[full_name]}",
#     "membership": "{sender[login]} {action} member {member[login]} to team " "{team[name]} in {repository[full_name]}",
#     "page_build": "{sender[login]} built pages in {repository[full_name]}",
#     "ping": "ping from {sender[login]}",
#     "public": "{sender[login]} publicized {repository[full_name]}",
#     "pull_request": "{sender[login]} {action} pull #{pull_request[number]} in " "{repository[full_name]}",
#     "pull_request_review": "{sender[login]} {action} {review[state]} "
#     "review on pull #{pull_request[number]} in "
#     "{repository[full_name]}",
#     "pull_request_review_comment": "{comment[user][login]} {action} comment "
#     "on pull #{pull_request[number]} in "
#     "{repository[full_name]}",
#     "push": "{pusher[name]} pushed {ref} in {repository[full_name]}",
#     "release": "{release[author][login]} {action} {release[tag_name]} in " "{repository[full_name]}",
#     "repository": "{sender[login]} {action} repository " "{repository[full_name]}",
#     "status": "{sender[login]} set {sha} status to {state} in " "{repository[full_name]}",
#     "team_add": "{sender[login]} added repository {repository[full_name]} to " "team {team[name]}",
#     "watch": "{sender[login]} {action} watch in repository " "{repository[full_name]}",
# }
#
#
# def _format_event(event_type, data):
#     try:
#         return EVENT_DESCRIPTIONS[event_type].format(**data)
#     except KeyError:
#         return event_type



from .matcher import (
    Matcher,
    MatcherOptions,
    MatcherOption
)
from .contact_matcher import ContactMatcher
from .room_matcher import RoomMatcher
from .message_matcher import MessageMatcher

__all__ = [
    'ContactMatcher',
    'RoomMatcher',
    'MessageMatcher',

    'Matcher',
    'MatcherOption',
    'MatcherOptions',
]



"""daily plugin"""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Union, List, Any

from apscheduler.schedulers.asyncio import AsyncIOScheduler     # type: ignore
from apscheduler.schedulers.base import BaseScheduler   # type: ignore

from wechaty import Wechaty, get_logger, Room, Contact      # type: ignore
from wechaty.plugin import WechatyPlugin, WechatyPluginOptions  # type: ignore


log = get_logger('DailyPlugin')


@dataclass
class DailyPluginOptions(WechatyPluginOptions):
    """daily plugin options"""
    room_id: Optional[str] = None
    contact_id: Optional[str] = None
    trigger: Optional[str] = None
    kwargs: Optional[dict] = None
    msg: Optional[str] = None


class DailyPlugin(WechatyPlugin):
    """
    say something everyday, like `Daily Words`
    """
    def __init__(self, options: DailyPluginOptions):
        super().__init__(options)

        if options.room_id is None and options.contact_id is None:
            raise Exception('room_id and contact_id should not all be empty')
        if options.room_id is not None and options.contact_id is not None:
            raise Exception('only one of the room_id contact_id should not '
                            'be none')
        if options.trigger is None:
            raise Exception('trigger should not be none')
        if options.kwargs is None:
            raise Exception('kwargs should not be none')
        if options.msg is None:
            raise Exception('msg should not be none')

        self.options: DailyPluginOptions = options
        self.scheduler: BaseScheduler = AsyncIOScheduler()
        self._scheduler_jobs: List[Any] = []

    @property
    def name(self) -> str:
        """get the name of the plugin"""
        if self.options is None or self.options.name is None:
            return 'dayily'
        return self.options.name

    async def tick(self, msg: str):
        """tick the things """
        print(msg)
        conversation: Union[Room, Contact]
        if self.options.room_id is not None:
            conversation = self.bot.Room.load(self.options.room_id)
        elif self.options.contact_id is not None:
            conversation = self.bot.Contact.load(self.options.contact_id)
        else:
            raise Exception('room_id and contact_id should not all be empty')

        await conversation.ready()
        await conversation.say(msg)

    async def init_plugin(self, wechaty: Wechaty):
        """init plugin"""
        await super().init_plugin(wechaty)
        for job in self._scheduler_jobs:
            job(wechaty)
        self.scheduler.start()

    def add_interval_job(self, func):
        """add interval job"""

        def add_job(bot: Wechaty):
            self.scheduler.add_job(
                func,
                trigger='interval',
                seconds=5,
                kwargs={
                    'bot': bot
                }
            )
        self._scheduler_jobs.append(add_job)



"""import all contrib"""
from wechaty_plugin_contrib.ding_dong_plugin import DingDongPlugin
from wechaty_plugin_contrib.daily_plugin import (
    DailyPluginOptions,
    DailyPlugin
)
from wechaty_plugin_contrib.rasa_rest_plugin import (
    RasaRestPluginOptions,
    RasaRestPlugin
)

from wechaty_plugin_contrib.contrib import (
    AutoReplyRule,
    AutoReplyOptions,
    AutoReplyPlugin
)

__all__ = [
    'DingDongPlugin',

    'DailyPluginOptions',
    'DailyPlugin',

    'RasaRestPluginOptions',
    'RasaRestPlugin',

    'AutoReplyRule',
    'AutoReplyOptions',
    'AutoReplyPlugin'
]



"""Room Matcher to match the specific room"""
import re
from re import Pattern
import inspect
from typing import List

from wechaty import Wechaty # type: ignore

from wechaty_plugin_contrib.config import (
    get_logger,
    Room,
)

from .matcher import Matcher


logger = get_logger("RoomMatcher")


class RoomMatcher(Matcher):
    async def match(self, target: Room) -> bool:
        """match the room"""
        logger.info(f'RoomMatcher match({target})')

        if not isinstance(target, Room):
            return False

        for option in self.options:

            if isinstance(option, Pattern):
                # match the room with regex pattern
                topic = await target.topic()
                is_match = re.match(option, topic) is not None
            elif isinstance(option, str):
                topic = await target.topic()
                is_match = target.room_id == option or topic == option

            # TODO: support check callback
            # elif hasattr(option, '__call__'):
            #     """check the type of the function
            #     refer: https://stackoverflow.com/a/56240578/6894382
            #     """
            #     if not inspect.isfunction(option):
            #         continue
            #     if inspect.iscoroutinefunction(option):
            #         is_match = await option(target)
            #     else:
            #         is_match = option(target)
            #
            #     if not isinstance(is_match, bool):
            #         raise ValueError('the type of result RoomMatcher function must be bool')

            elif isinstance(option, bool):
                return option

            else:
                raise ValueError(f'unknown type option: {option}')

            if is_match:
                return True
        return False

    async def find_rooms(self, bot: Wechaty) -> List[Room]:
        """find the matched rooms"""
        rooms = await bot.Room.find_all()
        matched_rooms: List[Room] = []
        for room in rooms:
            await room.ready()
            is_match: bool = await self.match(room)
            if is_match:
                matched_rooms.append(room)
        return matched_rooms



"""import basic config from wechaty-puppet"""

from wechaty import (   # type: ignore
    Room,
    Contact,
    Message,

    Wechaty
)

from wechaty_puppet import get_logger   # type: ignore

from .version import version

__all__ = [
    'get_logger',
    'version',

    'Room',
    'Contact',
    'Message',

    'Wechaty'
]



"""get the version from VERSION file"""
import os


def _get_version():
    file_path = os.path.join('../../', 'VERSION')
    if not os.path.exists(file_path):
        return '0.0.0'
    with open(file_path, 'r', encoding='utf-8') as f:
        return f.read().strip('\n')


version = _get_version()



"""
plugin validator
"""
from __future__ import annotations
from wechaty import WechatyPlugin   # type: ignore


def validate_plugin(plugin: WechatyPlugin):
    """validate the plugin"""
    # check the name of the plugin
    if isinstance(plugin, WechatyPlugin):
        raise Exception('plugin instance should a subclass of WechatyPlugin')

    # check the name of plugin
    if not hasattr(plugin, 'name'):
        raise Exception('WechatyPlugin must have name property')

    class_name = type(plugin).__name__
    if not class_name.endswith('Plugin'):
        raise Exception('the name of plugin <%s> should follow the Rule: '
                        'NamePlugin' % class_name)



"""Message Finder to match the specific message"""
import re
from re import Pattern
import inspect
from typing import List

from wechaty_plugin_contrib.config import (
    get_logger,
    Contact,
    Wechaty
)

from .finder import Finder


logger = get_logger("MessageFinder")


class MessageFinder(Finder):
    async def match(self, wechaty: Wechaty) -> List[Contact]:
        """match the room"""
        logger.info(f'MessageFinder match({Wechaty})')

        contacts: List[Contact] = []

        for option in self.options:
            if isinstance(option, Pattern):

                # search from all of the friends
                re_pattern = re.compile(option)
                # match the room with regex pattern
                all_friends = await wechaty.Contact.find_all()
                for friend in all_friends:
                    alias = await friend.alias()
                    if re.match(re_pattern, friend.name) or re.match(re_pattern, alias):
                        contacts.append(friend)

            elif isinstance(option, str):
                contact = wechaty.Contact.load(option)
                await contact.ready()
                contacts.append(contact)
            elif hasattr(option, '__call__'):
                """check the type of the function
                refer: https://stackoverflow.com/a/56240578/6894382
                """
                if inspect.iscoroutinefunction(option):
                    # pytype: disable=bad-return-type
                    targets = await option(wechaty)
                else:
                    targets = option(wechaty)

                if isinstance(targets, List[Contact]):
                    contacts.extend(targets)
            else:
                raise ValueError(f'unknown type option: {option}')
        return contacts



"""Base matcher for finding"""
from re import Pattern
from typing import (
    Callable,
    Optional,
    Union,
    List
)

from wechaty_plugin_contrib.config import (
    Room,
    Contact,
    Message
)


MatcherOption = Union[str, Pattern, bool, Callable[[Union[Contact, Room, Message]], bool]]
MatcherOptions = Union[MatcherOption, List[MatcherOption]]


class Matcher:
    def __init__(self, option: MatcherOptions):
        if isinstance(option, list):
            self.options = option
        else:
            self.options = [option]

    async def match(self, target) -> bool:
        raise NotImplementedError



"""exception in wechaty-plugin-contrib"""


class WechatyPluginError(Exception):
    """ Wechaty puppet error """

    def __init__(self, message, code=None, params=None):
        super().__init__(message, code, params)

        self.message = message
        self.code = code
        self.params = params

    def __str__(self):
        return repr(self)


class WechatyPluginConfigurationError(WechatyPluginError):
    """ Raises when configuration out of expected case """



from .auto_reply_plugin import (
    AutoReplyRule,
    AutoReplyOptions,
    AutoReplyPlugin
)
from .room_inviter import (
    RoomInviterPlugin,
    RoomInviterOptions
)

__all__ = [
    'AutoReplyOptions',
    'AutoReplyPlugin',
    'AutoReplyRule',

    'RoomInviterPlugin',
    'RoomInviterOptions'
]



"""Message Matcher to match the specific message"""
import re
from re import Pattern
import inspect
from typing import Awaitable

from wechaty_plugin_contrib.config import (
    get_logger,
    Message,
)

from .matcher import Matcher


logger = get_logger("MessageMatcher")


class MessageMatcher(Matcher):
    async def match(self, target: Message) -> bool:
        """match the room"""
        logger.info(f'MessageMatcher match({target})')

        if not isinstance(target, Message):
            return False

        for option in self.options:
            if isinstance(option, Pattern):
                # match the room with regex pattern
                topic = target.text()
                is_match = re.match(option, topic) is not None
            elif isinstance(option, str):
                await target.ready()
                is_match = target.message_id == option or target.text() == option

            # TODO: support check callback
            # elif callable(option):
            #     """check the type of the function
            #     refer: https://stackoverflow.com/a/56240578/6894382
            #     """
            #     if inspect.iscoroutinefunction(option):
            #         is_match = await option(target)
            #     else:
            #         is_match = option(target)

            elif isinstance(option, bool):
                return option

            else:
                raise ValueError(f'unknown type option: {option}')

            if is_match:
                return True
        return False



"""AutoReply to someone according to keywords"""
from dataclasses import dataclass, field
from typing import Union, List, Dict

from wechaty import (  # type: ignore
    WechatyPlugin,
    WechatyPluginOptions,
    FileBox,
    Contact,
    Message,
    Room,
    MiniProgram
)

from wechaty_puppet import (    # type: ignore
    get_logger,
)

from wechaty_plugin_contrib.matchers import (
    Matcher,
    MatcherOptions,
    MatcherOption, RoomMatcher, MessageMatcher, ContactMatcher
)

from wechaty_plugin_contrib.exception import (
    WechatyPluginConfigurationError
)


@dataclass
class AutoReplyRule:
    keyword: str
    reply_content: Union[str, FileBox, Contact, Message, MiniProgram]


@dataclass
class AutoReplyOptions(WechatyPluginOptions):
    rules: List[AutoReplyRule] = field(default_factory=list)
    matchers: List[Matcher] = field(default_factory=list)

    # add rules to the specific Matcher
    matcher_rules: Dict[Matcher, List[AutoReplyRule]] = field(default_factory=dict)


logger = get_logger('AutoReplyPlugin')


class AutoReplyPlugin(WechatyPlugin):

    def __init__(self, options: AutoReplyOptions):
        super().__init__(options)

        if not options.rules and not options.matcher_rules.values():
            raise WechatyPluginConfigurationError('rules not found in rules/matcher_rules')

        # if there is no matcher, it should be apply to all of target
        if not options.matchers:
            options.matchers = [RoomMatcher(True), MessageMatcher(True), ContactMatcher(True)]

        self.rule_map: Dict[Matcher, List[AutoReplyRule]] = {}
        # 1. build rules in common data
        for rule in options.rules:
            for matcher in options.matchers:
                if matcher not in self.rule_map:
                    self.rule_map[matcher] = []
                self.rule_map[matcher].append(rule)

        # 2. build the specific rules on matcher
        for matcher, rules in options.matcher_rules.items():
            if matcher not in self.rule_map:
                self.rule_map[matcher] = []
            self.rule_map[matcher].extend(rules)

    @property
    def matcher_options(self) -> List[Matcher]:
        if not isinstance(self.options, AutoReplyOptions):
            raise TypeError(f'{self} is not AutoReplyOptions type')
        return self.options.matchers

    async def is_match(self, target: Union[Contact, Message, Room]):
        """if there is only one rule which match the target, return True"""
        for option in self.matcher_options:
            _is_match = await option.match(target)
            if _is_match:
                return True
        return False

    async def on_message(self, msg: Message):
        """check the keyword and reply to talker"""

        conversation: Union[Room, Contact] = msg.room() if msg.room() else msg.talker()
        await conversation.ready()

        text = msg.text()

        # find the match rules
        for matcher, rules in self.rule_map.items():
            is_match = await matcher.match(conversation)

            if is_match:
                for rule in rules:
                    if rule.keyword == text:
                        await conversation.say(rule.reply_content)
